import argparse
from oracle import oracle_calculator
from web3 import Web3


marketplaces = ['OpenSea', 'LooksRare', 'X2Y2']


def check_hour(value):
    ivalue = int(value)
    if ivalue < 12 or ivalue > 720:
        raise argparse.ArgumentTypeError("%s is an invalid positive int value. Minimum is 12 hours, and maximum is 720 hours (30 days)." % value)
    return ivalue


def check_market(market):
    if market not in marketplaces:
        raise argparse.ArgumentTypeError("%s is not a supported market." % market)
    return market


def check_contract(contract):
    if not Web3.isAddress(contract):
        raise argparse.ArgumentTypeError("%s is not a valid contract address." % contract)
    return contract


parser = argparse.ArgumentParser(description='Description of your program')
parser.add_argument('-contract','--contract', type=check_contract, help='Contract address', required=True)
parser.add_argument('-markets','--marketplaces', type=check_market, nargs='+', help='Marketplaces to query', required=True)
parser.add_argument('-start','--start_hours', type=check_hour, help='Time window start hours', required=True)
parser.add_argument('-end','--end_hours', type=check_hour, help='Time window end hours', required=True)
args = vars(parser.parse_args())

start = int(args['start_hours'])
end = int(args['end_hours'])

print('Contract: {}'.format(args['contract']))
print('Marketplaces: {}'.format(args['marketplaces']))
print('Start hours: {}'.format(start))
print('End hours: {}'.format(end))

price = oracle_calculator(args['contract'], args['marketplaces'], start, end)

print('\n----------------------')
print('Oracle price: {}'.format(price[0]))
print('Oracle volatility: {}'.format(price[1]))
print('----------------------')

# python main.py --contract '0xbc4ca0eda7647a8ab7c2061c2e118a18a936f13d' --marketplaces 'OpenSea' 'LooksRare' --start_hours 720 --end_hours 12


import requests
from decouple import config
import datetime
import dateparser
import time
from processors import processOpensea, processLooksRare, processX2Y2
import random
from concurrent.futures import ThreadPoolExecutor


headers = {
    'opensea':
        {
            "Accept": "application/json",
            "X-API-KEY": config("OPENSEA")
        },
    'x2y2':
        {
            "Accept": "application/json",
            "X-API-KEY": config("X2Y2")
        },
    }

urls = {'opensea': "https://api.opensea.io/api/v1/events?&asset_contract_address={}&event_type=successful&occurred_before={}&occurred_after={}",

        'looksrare': "https://api.looksrare.org/api/v1/events?collection={}&type=SALE&pagination%5Bfirst%5D=150",
        
        'x2y2': "https://api.x2y2.org/v1/events?type=sale&contract={}&created_before={}&created_after={}",
        }

loans = {}


def requestOpenSea(contract, start_hours, end_hours):
    # print('Requesting OpenSea...')
    prices = []
    tokens = []
    pages = 0
    
    start_date = time.mktime((dateparser.parse(str(start_hours) + " hours ago")).timetuple())
    end_date = time.mktime((dateparser.parse(str(end_hours) + " hours ago")).timetuple())

    opensea_data = urls["opensea"].format(contract, end_date, start_date)
    print(opensea_data)
    entries = requests.get(opensea_data, headers=headers["opensea"]).json()
    tokens, prices = processOpensea(entries, tokens)

    while entries["next"]:
        print('Next OpenSea pagination...')
        time.sleep(1)
        # Stop when volume is a lot
        if len(prices) > 40:
            print('High volume reached')
            return prices
        else:
            entries = requests.get(opensea_data + "&cursor=" + entries['next'], headers=headers["opensea"]).json()
            tokens_return, prices_return = processOpensea(entries, tokens)
            tokens = tokens + tokens_return
            prices = prices + prices_return

    return prices


def requestLooksRare(contract, start_hours, end_hours):
    # print('Requesting LooksRare...')
    prices = []
    tokens = []

    start_date = datetime.datetime.strptime(str(dateparser.parse(str(start_hours) + " hours ago"))[:-7], '%Y-%m-%d %H:%M:%S')
    end_date = datetime.datetime.strptime(str(dateparser.parse(str(end_hours) + " hours ago"))[:-7], '%Y-%m-%d %H:%M:%S')
    
    entries = requests.get(urls["looksrare"].format(contract)).json()
    tokens, prices, state = processLooksRare(entries, tokens, start_date, end_date)

    # Finished searching
    if state:
        pass
    else:
        while True:
            print('Next LooksRare pagination...')
            entries = requests.get(urls["looksrare"].format(contract) + "&pagination%5Bcursor%5D=" + state).json()
            tokens_return, prices_return, state = processLooksRare(entries, tokens, start_date, end_date)
            tokens = tokens + tokens_return
            prices = prices + prices_return

    return prices


def requestX2Y2(contract, start_hours, end_hours):
    # print('Requesting X2Y2...')
    prices = []
    tokens = []

    start_date = round(time.mktime((dateparser.parse(str(start_hours) + " hours ago")).timetuple()))
    end_date = round(time.mktime((dateparser.parse(str(end_hours) + " hours ago")).timetuple()))

    x2y2_data = urls["x2y2"].format(contract, end_date, start_date)
    entries = requests.get(x2y2_data, headers=headers["x2y2"]).json()
    tokens, prices = processX2Y2(entries, tokens)

    while entries["next"]:
        time.sleep(1)
        entries = requests.get(x2y2_data + "&cursor=" + entries["next"], headers=headers["x2y2"]).json()
        tokens_return, prices_return = processX2Y2(entries, tokens)
        tokens = tokens + tokens_return
        prices = prices + prices_return

    return prices


def poolParser(args):
    if 'OpenSea' == args[0]:
        overall_prices = requestOpenSea(args[1], args[2], args[3])
    elif 'LooksRare' == args[0]:
        overall_prices = requestLooksRare(args[1], args[2], args[3])
    elif 'X2Y2' == args[0]:
        overall_prices = requestX2Y2(args[1], args[2], args[3])
    
    return overall_prices


def fetchData(contract, marketplaces, start, end):
    overall_prices = []
    args = []
    print('Fetching data for contract {} in marketplaces: {}'.format(contract, marketplaces))

    for market in marketplaces:
        args.append((market, contract, start, end))

    with ThreadPoolExecutor(max_workers=3) as pool:
        response_list = list(pool.map(poolParser, args))

    overall_prices = [int(val) for sublist in response_list for val in sublist]

    # Minimal threshold for data entries
    if len(overall_prices) < 10:
        print('Not enough data entries for oracle calculation. Please try different time windows')
        return 'error'

    return overall_prices


def average_list(list):
    avg = sum(list) / len(list) 
    return avg


def random_range(end, start):
    while True:
        num_list = random.sample(range(end, start), 2)
        random_difference = abs(num_list[0] - num_list[1])
        if random_difference == round((start - end) * 0.8):
            num_list.sort()
            random_end = num_list[0]
            random_start = num_list[1]

            print('Random time window: {}'.format(num_list))
            print('Random time difference: {}'.format(random_difference))
            return random_end, random_start



from helpers import fetchData, average_list, random_range
import numpy as np
import os
import csv
from datetime import datetime
import math


def oracle_calculator(contract, marketplaces, start, end):
    file = './cache/{}.csv'.format(contract)

    # Check if cache file exists
    if os.path.isfile(file):
        with open(file) as f:
            reader = csv.reader(f)
            for row in reader:
                date = datetime.strptime(row[0], '%Y-%m-%d %H:%M:%S.%f')
                final_price = int(row[1])
                volatility = float(row[2])

        # Check if cached data is one day old
        if (datetime.now() - date).days > 1:
            # 24 hours have passed
            pass
        else:
            # Date is within 24 hours
            return final_price, volatility

    end_hours, start_hours = random_range(end, start)
    price_entries = fetchData(contract, marketplaces, start_hours, end_hours)

    if isinstance(price_entries, str):
        # Not enough volume
        return 0, 0

    print(price_entries)

    log_returns = [math.log(number) for number in price_entries]
    volatility = np.std(log_returns) * 100

    print('Received query list:')
    print(price_entries)
    print('\nData filtering...\n')

    average = average_list(price_entries)

    # Data filtering
    for price in price_entries:
        if price < 0.1 * average or price > 9 * average:
            print('Removing price {} as it is a extremely low or high value'.format(price))
            price_entries.remove(price)
    
    price_entries.sort()

    # Truncated average with 10%
    number_to_remove = round(len(price_entries) * 0.1)
    price_entries = price_entries[number_to_remove:-number_to_remove]

    truncated_average = average_list(price_entries)
    truncated_std = np.std(price_entries)

    print("Truncated standard deviation of list: % s "% (truncated_std))
    print("Truncated average of list: {}\n".format(truncated_average))

    prices = []
    sensitivity = 1

    for price in price_entries:
        if truncated_average - (sensitivity * truncated_std) <= price <= truncated_average + (sensitivity * truncated_std):
            prices.append(price)

    print('Final price list: ')
    print(prices)

    final_price = int(np.format_float_positional(average_list(prices), trim='-'))

    if not os.path.exists("./cache"):
        # Create a new directory because it does not exist
        os.makedirs("./cache")

    f = open(file, "w+")
    f.write("{},{},{}".format(datetime.now(), final_price, volatility))
    f.close()

    return final_price, volatility



from datetime import datetime


def processOpensea(entries, stored):
    # print('Stored tokens: ' + str(stored))
    prices = []
    tokens = []

    for entry in entries['asset_events']:
        try:
            # print('Token ID: {}'.format(entry['asset']['token_id']))
            # print(entry['is_private'])
            # print(entry['quantity'] == '1')
            # print(entry['payment_token']['symbol'])
            # check that its sold in eth, quantity one, is_private == false and id is not seen before
            if entry['is_private'] == False and entry['quantity'] == '1' and entry['payment_token']['symbol'] == 'ETH' and entry['asset']['token_id'] not in stored:
                # print('Price  (Wei): {}'.format(entry['total_price']))
                prices.append(entry['total_price'])
                tokens.append(entry['asset']['token_id'])
        except Exception as e:
            print(e)
            continue

    return tokens, prices


def processLooksRare(entries, stored, start, end):
    # print('Stored tokens: ' + str(stored))
    prices = []
    tokens = []
    next_page = ''

    for entry in entries['data']:
        if entry['order']['amount'] == 1 and entry['order']['currencyAddress'] == '0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2' and entry['token']['tokenId'] not in stored:
            date = datetime.strptime((entry['createdAt'].replace('T', ' '))[:-5], '%Y-%m-%d %H:%M:%S')
            if start <= date <= end:
                # print()
                # print('Token ID: {}'.format(entry['token']['tokenId']))
                # print('Price  (Wei): {}'.format(entry['order']['price']))
                # print('Date: {}'.format(date))
                # print()

                next_page = entry['id']
                prices.append(entry['order']['price'])
                tokens.append(entry['token']['tokenId'])
            else:
                # print('Reached end with date: ' + str(date))
                return tokens, prices, True

    return tokens, prices, next_page


def processX2Y2(entries, stored):
    # print('Stored tokens: ' + str(stored))
    prices = []
    tokens = []

    for entry in entries['data']:
        # check that its sold in eth, quantity one, bundle == false and id is not seen before
        if entry['order']['amount'] == 1 and entry['order']['currency'] == '0x0000000000000000000000000000000000000000' and entry['order']['is_bundle'] == False and entry['token']['token_id'] not in stored:
            # print()
            # print('Token ID: {}'.format(entry['token']['token_id']))
            # print('Price  (Wei): {}'.format(entry['order']['price']))
            prices.append(entry['order']['price'])
            tokens.append(entry['token']['token_id'])

    return tokens, prices